import 'dart:convert';
import 'dart:ffi';
import 'package:solana_rpc_sdk_flutter/helper/cluster.dart';
import 'package:solana_rpc_sdk_flutter/helper/rpc_method.dart';
import 'package:solana_rpc_sdk_flutter/helper/token.dart';
import 'package:solana_rpc_sdk_flutter/helper/http_client.dart';
import 'package:solana_rpc_sdk_flutter/model/get_account_Info.dart';
import 'package:solana_rpc_sdk_flutter/model/get_balance.dart';
import 'package:solana_rpc_sdk_flutter/model/get_supply.dart';
import 'package:solana_rpc_sdk_flutter/model/get_transaction.dart';
import 'package:solana_rpc_sdk_flutter/model/get_version.dart';
import 'package:solana_rpc_sdk_flutter/model/get_vote_accounts.dart';
import 'package:solana_rpc_sdk_flutter/model/rpc_response.dart';
import 'package:solana_rpc_sdk_flutter/model/token_accounts_by_owner.dart';

import '../model/get_token_account_balance.dart';

/// A Calculator.
class SolanaRpcApi {
  final String _url;
  final String address;
  final RPCHttpClient _httpClient;

  /// constructs
  SolanaRpcApi(this._url, this.address) : _httpClient = RPCHttpClient();

  /// get TokenAccount Balance By Pubkey address
  Future<RPCResponse<GetTokenAccountBalance>> getTokenAccountBalance(
      String pubkey) async {
    final body = json.encode({
      "jsonrpc": "2.0",
      "id": 1,
      "method": RPCMethod.getTokenAccountBalance,
      "params": [
        pubkey,
      ]
    });
    final response = await _httpClient.post(_url, data: body);
    return RPCResponse<GetTokenAccountBalance>.fromJson(response.data,
        onData: (Map<String, dynamic> value) {
      return GetTokenAccountBalance.fromJson(value);
    });
  }

  /// get AccountInfo
  Future<RPCResponse<GetAccountInfo>> getAccountInfo({String? pubkey}) async {
    final body = json.encode({
      "jsonrpc": "2.0",
      "id": 1,
      "method": RPCMethod.getAccountInfo,
      "params": [
        pubkey ?? address,
        {"encoding": "jsonParsed"}
      ]
    });
    final response = await _httpClient.post(_url, data: body);
    return RPCResponse<GetAccountInfo>.fromJson(response.data,
        onData: (Map<String, dynamic> value) {
      return GetAccountInfo.fromJson(value);
    });
  }

  /// get Balance
  Future<RPCResponse<GetBalance>> getBalance({String? pubkey}) async {
    final body = json.encode({
      "jsonrpc": "2.0",
      "id": 1,
      "method": RPCMethod.getBalance,
      "params": [
        pubkey ?? address,
      ]
    });
    final response = await _httpClient.post(_url, data: body);
    return RPCResponse<GetBalance>.fromJson(response.data,
        onData: (Map<String, dynamic> value) {
      return GetBalance.fromJson(value);
    });
  }

  /// get Supply
  Future<RPCResponse<GetSupply>> getSupply() async {
    final body = json.encode({
      "jsonrpc": "2.0",
      "id": 1,
      "method": RPCMethod.getSupply,
    });
    final response = await _httpClient.post(_url, data: body);
    return RPCResponse<GetSupply>.fromJson(response.data,
        onData: (Map<String, dynamic> value) {
      return GetSupply.fromJson(value);
    });
  }

  /// get TokenAccounts By Owner with [mint] or [token program]
  Future<RPCResponse<TokenAccountsByOwner>> getTokenAccountsByOwner(
      {String? mint}) async {
    final body = json.encode({
      "jsonrpc": "2.0",
      "id": 1,
      "method": RPCMethod.getTokenAccountsByOwner,
      "params": [
        address,
        mint != null ? {'mint: $mint'} : {"programId": Token.TOKEN_PROGRAM_ID},
        {"encoding": "jsonParsed"}
      ]
    });
    final response = await _httpClient.post(_url, data: body);
    return RPCResponse<TokenAccountsByOwner>.fromJson(response.data,
        onData: (Map<String, dynamic> value) {
      return TokenAccountsByOwner.fromJson(value);
    });
  }

  /// get Transaction detail by [hash]
  Future<RPCResponse<GetTransaction>> getTransaction(String hash) async {
    final body = json.encode({
      "jsonrpc": "2.0",
      "id": 1,
      "method": RPCMethod.getTransaction,
      "params": [
        hash,
        {"encoding": "jsonParsed"}
      ]
    });
    final response = await _httpClient.post(_url, data: body);

    return RPCResponse<GetTransaction>.fromJson(
      response.data,
      onData: (Map<String, dynamic> value) {
        return GetTransaction.fromJson(value);
      },
    );
  }

  /// get Transaction Count
  Future<RPCResponse<int>> getTransactionCount() async {
    final body = json.encode({
      "jsonrpc": "2.0",
      "id": 1,
      "method": RPCMethod.getTransactionCount,
    });
    final response = await _httpClient.post(_url, data: body);

    return RPCResponse<int>.fromJson(
      response.data,
    );
  }

  /// get Version by rpc node
  Future<RPCResponse<GetVersion>> getVersion() async {
    final body = json.encode({
      "jsonrpc": "2.0",
      "id": 1,
      "method": RPCMethod.getVersion,
    });
    final response = await _httpClient.post(_url, data: body);

    return RPCResponse<GetVersion>.fromJson(
      response.data,
      onData: (Map<String, dynamic> value) {
        return GetVersion.fromJson(value);
      },
    );
  }

  /// get Vote Accounts
  Future<RPCResponse<GetVoteAccounts>> getVoteAccounts({
    String? votePubkey,
    bool? keepUnstakedDelinquents,
    int? delinquentSlotDistance,
  }) async {
    final params = [];

    if (votePubkey != null) {
      params.add({"votePubkey": votePubkey});
    }

    if (keepUnstakedDelinquents != null) {
      params.add({"keepUnstakedDelinquents": keepUnstakedDelinquents});
    }

    if (delinquentSlotDistance != null) {
      params.add({"delinquentSlotDistance": delinquentSlotDistance});
    }

    final body = json.encode({
      "jsonrpc": "2.0",
      "id": 1,
      "method": RPCMethod.getVoteAccounts,
      "params": params
    });
    final response = await _httpClient.post(_url, data: body);

    return RPCResponse<GetVoteAccounts>.fromJson(
      response.data,
      onData: (Map<String, dynamic> value) {
        return GetVoteAccounts.fromJson(value);
      },
    );
  }

  /// minimum Ledger Slot
  Future<RPCResponse<int>> minimumLedgerSlot() async {
    final body = json.encode({
      "jsonrpc": "2.0",
      "id": 1,
      "method": RPCMethod.minimumLedgerSlot,
    });
    final response = await _httpClient.post(_url, data: body);

    return RPCResponse<int>.fromJson(
      response.data,
    );
  }

  /// request Airdrop
  Future<RPCResponse<String>> requestAirdrop(
      {required String pubkey, required int amount}) async {
    if (_url == clusterApiUrl(Cluster.mainnetBeta)) {
      return Future.error(Exception('Not supported'));
    }

    final body = json.encode({
      "jsonrpc": "2.0",
      "id": 1,
      "method": RPCMethod.requestAirdrop,
      "params": [pubkey, amount]
    });
    final response = await _httpClient.post(_url, data: body);

    return RPCResponse<String>.fromJson(
      response.data,
    );
  }
}
